Package: BiocNeighbors
Version: 1.22.0
Date: 2022-12-17
Title: Nearest Neighbor Detection for Bioconductor Packages
Authors@R: c(person("Aaron", "Lun", role=c("aut", "cre", "cph"),
    email="infinite.monkeys.with.keyboards@gmail.com"))
Imports: Rcpp, S4Vectors, BiocParallel, stats, methods, Matrix
Suggests: testthat, BiocStyle, knitr, rmarkdown, FNN, RcppAnnoy,
        RcppHNSW
biocViews: Clustering, Classification
Description: Implements exact and approximate methods for nearest neighbor 
    detection, in a framework that allows them to be easily switched within 
    Bioconductor packages or workflows. Exact searches can be performed using
    the k-means for k-nearest neighbors algorithm or with vantage point trees.
    Approximate searches can be performed using the Annoy or HNSW libraries.
    Searching on either Euclidean or Manhattan distances is supported. 
    Parallelization is achieved for all methods by using BiocParallel. Functions 
    are also provided to search for all neighbors within a given distance. 
License: GPL-3
LinkingTo: Rcpp, RcppHNSW
VignetteBuilder: knitr
SystemRequirements: C++11
RoxygenNote: 7.1.1
git_url: https://git.bioconductor.org/packages/BiocNeighbors
git_branch: RELEASE_3_19
git_last_commit: c9f4480
git_last_commit_date: 2024-04-30
Repository: Bioconductor 3.19
Date/Publication: 2024-04-30
NeedsCompilation: yes
Packaged: 2024-05-01 02:47:49 UTC; biocbuild
Author: Aaron Lun [aut, cre, cph]
Maintainer: Aaron Lun <infinite.monkeys.with.keyboards@gmail.com>
Built: R 4.4.1; x86_64-apple-darwin24.0.0; 2024-10-14 04:18:42 UTC; unix
