# Generated by roxygen2: do not edit by hand

export(AnnoyIndex)
export(AnnoyIndex_path)
export(AnnoyIndex_search_mult)
export(AnnoyParam)
export(AnnoyParam_directory)
export(AnnoyParam_ntrees)
export(AnnoyParam_search_mult)
export(ExhaustiveIndex)
export(ExhaustiveParam)
export(HnswIndex)
export(HnswIndex_ef_search)
export(HnswIndex_path)
export(HnswParam)
export(HnswParam_directory)
export(HnswParam_ef_construction)
export(HnswParam_ef_search)
export(HnswParam_nlinks)
export(KmknnIndex)
export(KmknnIndex_cluster_centers)
export(KmknnIndex_cluster_info)
export(KmknnParam)
export(KmknnParam_kmeans_args)
export(VptreeIndex)
export(VptreeIndex_nodes)
export(VptreeParam)
export(bndata)
export(bndistance)
export(bnorder)
export(buildAnnoy)
export(buildExhaustive)
export(buildHnsw)
export(buildIndex)
export(buildKmknn)
export(buildVptree)
export(findAnnoy)
export(findExhaustive)
export(findHnsw)
export(findKNN)
export(findKmknn)
export(findMutualNN)
export(findNeighbors)
export(findVptree)
export(queryAnnoy)
export(queryExhaustive)
export(queryHnsw)
export(queryKNN)
export(queryKmknn)
export(queryNeighbors)
export(queryVptree)
export(rangeFindExhaustive)
export(rangeFindKmknn)
export(rangeFindVptree)
export(rangeQueryExhaustive)
export(rangeQueryKmknn)
export(rangeQueryVptree)
exportClasses(AnnoyIndex)
exportClasses(AnnoyParam)
exportClasses(BiocNeighborIndex)
exportClasses(BiocNeighborParam)
exportClasses(ExhaustiveIndex)
exportClasses(ExhaustiveParam)
exportClasses(HnswIndex)
exportClasses(HnswParam)
exportClasses(KmknnIndex)
exportClasses(KmknnParam)
exportClasses(VptreeIndex)
exportClasses(VptreeParam)
exportMethods("[[")
exportMethods("[[<-")
exportMethods(bndata)
exportMethods(bndistance)
exportMethods(bnorder)
exportMethods(buildIndex)
exportMethods(dim)
exportMethods(dimnames)
exportMethods(findKNN)
exportMethods(findNeighbors)
exportMethods(queryKNN)
exportMethods(queryNeighbors)
exportMethods(show)
import(BiocParallel)
import(methods)
importClassesFrom(S4Vectors,character_OR_NULL)
importFrom(BiocParallel,SerialParam)
importFrom(BiocParallel,bpmapply)
importFrom(BiocParallel,bpnworkers)
importFrom(Matrix,t)
importFrom(Rcpp,sourceCpp)
importFrom(S4Vectors,setValidity2)
importFrom(methods,is)
importFrom(methods,new)
importFrom(methods,show)
importFrom(stats,kmeans)
useDynLib(BiocNeighbors)
